/* Finds the best Euclidean fit of two fragments by branch-and-bound */
/* Last edited on 2006-03-19 12:16:35 by stolfi */

#include <fbox.h>
#include <fboxheap.h>
#include <fboxlist.h>
#include <aa.h>
#include <ia.h>
#include <affirm.h>
#include <ps.h>
#include <stdlib.h>
#include <stdio.h>
#include <bbopt.h>
#include <bbgoal.h>

#define D 3

typedef struct Options 
  { char *curve1;         /* Name of first curve. */
    char *curve2;         /* Name of second curve. */
    char *outName;        /* Prefix for output files. */
    /* Plotting options: */
    bool_t epsfmt;          /* TRUE for encapsulated postscript output. */
    /* Parameters for branch-and-bound optimization: */
    double tol;           /* Positional accuracy required. */
    double sigma;         /* SD of noise in curve samples. */
  } Options;
  /* Command line options passed to the program. */

/* INTERNAL PROTOS */

int main(int argc, char **argv);

void cf_initialize_plot
  ( char *outName,
    bool_t epsfmt, 
    Interval xr, 
    Interval yr,
    FILE **psfP
  );
  /* Opens the output plot file and initializes scales, captions, etc.
    If {epsfmt} is true the file is encapsulated Postscript
    ("{outname}.eps") otherwise it is a plain Postscripr
    ("{outname}.ps"). The file is returned in {*psfP}. */

void cf_finalize_plot(FILE **psfP, bool_t epsfmt);
   /* Finalizes the plot and closes the plot file {*psfP}, whose
     type is indicated by {epsf}. */

void cf_print_box(Interval *xr, Interval *fr, bool_t final);
  /* A reporting function that prints the domain box {xr[0..D-1]} and 
    its image {*fr} to {stderr}. */

/* IMPLEMENTATIONS */

int main(int argc, char **argv)
  { Options *o = cf_parse_options(argc, argv);
    char *outName = txtcat(txtcat("cf-", o->curve1), txtcat("-", o->curve2));
    FILE *psfile = NULL;
    
    ia_init();

    /* Read the two curves, get barycenters and radii: */
    PointVec c1 = cf_read_curve(o->curve1);
    Point b1 = cf_barycenter(c1);
    double R1 = cf_radius(c1, b1);

    PointVec c2 = cf_read_curve(o->curve2);
    Point b2 = cf_barycenter(c2);
    double R2 = cf_radius(c2, b2);
    
    /* Choose the initial domain {xr[0..D-1]} and get its width {xw[0..D-1]}: */
    /*   xr[0] = relative X displacement of center of curve 2 rel to curve 1. */
    /*   xr[1] = rotation angle of curve 1. */
    /*   xr[2] = rotation angle of curve 2. */
    /* The angles xr[1],xr[2] are in radians times the mean radii. */
    Interval xr[D], tol[D];
    Float xw[D];
    xr[0] = (Interval){0, R1+R2};         tol[0] = o->tol;
    xr[1] = (Interval){0.0, 2*3.1416*R1}; tol[1] = o->tol;
    xr[2] = (Interval){0.0, 2*3.1416*R2}; tol[2] = o->tol;

    /* Initialize the output file: */
    cf_initialize_plot(o->outName, o->epsfmt, R1, R2, &psfile);

    /* Optimize and plot node boxes: */

    auto void draw_state(Interval *xr, Interval fr, bool_t final);
    void draw_state(Interval *xr, Interval fr, bool_t final)
      { cf_print_state(xr, &fr, final);
        Interval d = xr[0];
        Interval t1 = ia_scale(xr[1], 1.0, R1);
        Interval t2 = ia_scale(xr[2], 1.0, R2);
        nevals++;
        if ((nevals < 20) || (nevals % 20 == 0) || final)
          { cf_new_drawing(psfile, nevals);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.lo, t1.lo, t2.lo);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.lo, t1.lo, t2.hi);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.lo, t1.hi, t2.lo);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.lo, t1.hi, t2.hi);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.hi, t1.lo, t2.lo);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.hi, t1.lo, t2.hi);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.hi, t1.hi, t2.lo);
            cf_draw_curves(psfile, c1, b1, R1, c2, b2, R2, d.hi, t1.hi, t2.hi);
          }
      }

    ps_set_pen(psfile, 0.0,0.5,0.0, 0.10, 0.0,0.0);
    Interval gr;
    FBoxList R = bb_optimize(3, cf_eval_fit, &(xr[0]), &(tol[0]), &gr, draw_state); 
    affirm(R != NULL, "bb_optimize returned no boxes");
    fprintf(stderr, "minimum found = [%8.4f _ %8.4f]\n", gr.lo, gr.hi);

     /* Terminate plot: */
    cf_finalize_plot(&psfile, epsfmt);

    return 0;
  }

void cf_initialize_plot
  ( char *outName,
    bool_t epsfmt, 
    double totR,
    FILE **psfP
  )
  {
    FILE *psfile;
  
    /* Compute widths of {xr} and {yr}: */
    double Rmax = (R1 > R2 ? R1 : R2);
    Interval xr = (Interval){ -(R1+R2), +(R1+R2) };  double xw = xr.hi - xr.lo;
    Interval yr = (Interval){ -Rmax, +Rmax };        double yw = yr.hi - yr.lo;  
    
    /* Add a safety margin around {xr,yr}: */
    xr.lo -= 0.02*xw; xr.hi += 0.02*xw; xw = xr.hi - xr.lo;
    yr.lo -= 0.02*yw; yr.hi += 0.02*yw; yw = yr.hi - yr.lo;

    /* Choose device window in the same proportion: */
    double xydyag = hypot(xw, yw); /* Diagonal of client plot window. */
    double hvdiag = 600.0;         /* Approx. ideal diagonal of device window (pt). */
    
    double hw = ceil(xw * hvdiag/xydiag);
    double vw = ceil(yw * hvdiag/xydiag);
    
    if (epsfmt)
      { psfile = open_write(txtcat(outName, ".eps"), TRUE);
        double mrg = 4.0;
        ps_begin_figure(psfile, 0.0, hw + 2*mrg, 0.0, vw + 2*mrg);
        ps_set_window
          ( psfile,
            xr.lo, xr.hi,
            yr.lo, yr.hi,
            mrg, hw + mrg,  mrg, hw + mrg,
            1, 1
          );
        ps_set_pen(psfile, 0.0,0.0,0.0, 0.10, 0.0,0.0);
      }
    else
      { double hctr = ( 8.5 * 72.0)/2.0;
        double vctr = (11.0 * 72.0)/2.0;
        psfile = open_write(txtcat(outName, ".ps"), TRUE);
        ps_begin_document(psfile, "letter");
        ps_begin_page(psfile, "1");
        ps_set_window
          ( psfile,
            xr.lo, xr.hi,
            yr.lo, yr.hi,
            hctr - hw/2, hctr + hw/2, vctr - vw/2, vctr + vw/2,
            1, 1
          );
        ps_set_pen(psfile, 0.0,0.0,0.0, 0.10, 0.0,0.0);
      }
    *psfP = psfile;
  }
  
void cf_new_drawing(FILE *psfile, int nevals)
  {
  }

void cf_draw_curves
  ( FILE *psfile,
    Point_vec_t c1,
    Point b1,
    double R1,
    Point_vec_t c2,
    Point b2,
    double R2,
    double d,
    double t1,
    double t2
  )
  { /* Compute positions of centers of two curves: */
    double x1 = -R1*d/(R1+R2), x2 = R2*d/(R1+R2);
    /* Draw curve 1 in red: */
    ps_set_pen(psfile, 0.667,0.000,0.000, 0.10, 0.0,0.0);
    cf_draw_curve(c1, b1, x1, t1);
    /* Draw curve 2 in blue: */
    ps_set_pen(psfile, 0.000,0.167,1.000, 0.10, 0.0,0.0);
    cf_draw_curve(c2, b2, x2, t2);
  }

void cf_finalize_plot(FILE **psfP, bool_t epsfmt)  
  {
    FILE *psfile = *psfP;
    if (epsfmt)
      { ps_end_figure(psfile);
        fclose(psfile);
      }
    else
      { ps_end_page(psfile); 
        ps_end_document(psfile);
        fclose(psfile);
      }
  }

void cf_print_state(Interval *xr, Interval *fr, bool_t final)
  {
    ia_print(stderr, xr[0]); 
    fprintf(stderr, "  ");
    ia_print(stderr, xr[1]); 
    fprintf(stderr, "  ");
    ia_print(stderr, xr[2]); 
    fprintf(stderr, " = ");
    ia_print(stderr, *fr); 
    if (final) { fprintf(stderr, " *"); }
    fprintf(stderr, "\n");
  }
  
#define ARG_ERROR(Msg,Arg) arg_error((Msg),(Arg),argv[0],usage)
  
#define GET_NEXT_STRING(Var) get_next_arg_string(&(Var), &argn, argc, argv, usage)
#define GET_NEXT_DOUBLE(Var) get_next_arg_double(&(Var), &argn, argc, argv, usage)
#define GET_NEXT_INT(Var) get_next_arg_int(&(Var), &argn, argc, argv, usage)
  
Options *parse_options(int argc, char **argv)
   /*
     Parses the command line options, returns a record with their options. */
  {
    char* usage = "\n  [ -help ] -curves NAME NAME \\\n  { -eps | -ps } -tol NUM -sigma NUM -maxEvals NUM \\\n  -outName NAME";
    Options *o = (Options *)malloc(sizeof(Options));
    int argn;

    /* Defaults: */
    o->curve1 = NULL;       /* For now. */
    o->curve2 = NULL;       /* For now. */
    o->eps = FALSE;         /* For now. */
    o->tol = -1.0;          /* For now. */
    o->sigma = -1.0;        /* For now. */
    o->maxEvals = -1;       /* For now. */
    o->outName = NULL;      /* For now. */
    
    argn = 1;

    /* Scan command line options. */
    while ((argn < argc) && (argv[argn][0] == '-') && (argv[argn][1] != '\0'))
      {
        char *key = argv[argn];
        if ((key[0] == '-') && (key[1] == '-') && (key[2] != '\0')) { key++; }
        if (strcmp(key, "-help") == 0)
          { fprintf(stderr, "usage: %s %s\n", argv[0], usage); exit(0); }
        else if (strcmp(key, "-curves") == 0)
          { GET_NEXT_STRING(o->curve1);
            GET_NEXT_STRING(o->curve2);
          }
        else if (strcmp(key, "-tol") == 0)
          { GET_NEXT_DOUBLE(o->tol);
            if (o->tol <= 0.0)
              { ARG_ERROR("invalid \"-tol\"", ""); }
          }
        else if (strcmp(key, "-sigma") == 0)
          { GET_NEXT_DOUBLE(o->sigma);
            if (o->sigma <= 0.0)
              { ARG_ERROR("invalid \"-sigma\"", ""); }
          }
        else if (strcmp(key, "-ps") == 0)
          { o->epsfmt = FALSE;
          }
        else if (strcmp(key, "-eps") == 0)
          { o->epsfmt = TRUE;
          }
        else if (strcmp(key, "-maxEvals") == 0)
          { GET_NEXT_INT(o->maxEvals);
            if (o->maxEvals < 0)
              { ARG_ERROR("invalid \"-maxEvals\"", ""); }
          }
        else if (strcmp(key, "-outName") == 0)
          { GET_NEXT_STRING(o->outName);
          }
        else 
          { ARG_ERROR("unknown option", argv[argn]); }
        ++argn;
      }
    if (argn != argc) { ARG_ERROR("extraneous arguments", argv[argn]); }
    if (o->curve1 == NULL) { ARG_ERROR("must specify \"-curves\"", ""); }
    if (o->tol < 0.0) { ARG_ERROR("must specify \"-tol\"", ""); }
    if (o->sigma < 0.0) { ARG_ERROR("must specify \"-sigma\"", ""); }
    if (o->maxEvals < 0.0) { o->maxEvals = 1000; }
    if (o->outName == NULL) { ARG_ERROR("must specify \"-outName\"", ""); }
    return o;
  }

void get_next_arg_string(char **varp, int *argnp, int argc, char **argv, char *usage)
   /*
     Stores the next command line argument (as a string) into "*varp" */
  {
    int argn = *argnp;
    if (argn+1 >= argc)
      { ARG_ERROR("missing arg value", argv[argn]); }
    (*varp) = argv[argn+1];
    (*argnp) = argn+1;
  }
  
void get_next_arg_double(double *varp, int *argnp, int argc, char **argv, char *usage)
   /*
     Stores the next command line argument (as a double) into "*varp" */
  {
    int argn = *argnp;
    char *end;
    if (argn+1 >= argc)
      { ARG_ERROR("missing arg value", argv[argn]); }
    (*varp) = strtod(argv[argn+1], &end);
    if ((*end) != '\0') 
      { ARG_ERROR("invalid numeric argument", argv[argn+1]); }
    (*argnp) = argn+1;
  }

void get_next_arg_int(int *varp, int *argnp, int argc, char **argv, char *usage)
   /*
     Stores the next command line argument (as an int) into "*varp" */
  {
    int argn = *argnp;
    double v;
    char *end;
    if (argn+1 >= argc)
      { ARG_ERROR("missing arg value", argv[argn]); }
    v = strtod(argv[argn+1], &end);
    if ((*end) != '\0') 
      { ARG_ERROR("invalid numeric argument", argv[argn+1]); }
    (*varp) = (int)(v + 0.5);
    (*argnp) = argn+1;
  }

void arg_error(char *msg, char *arg, char *pname, char *usage)
   /*
     Prints "msg", "arg", the "usage" message, and exits.
     Handy while parsing the command line arguments. */
  {
    fprintf(stderr, "%s %s\n", msg, arg);
    fprintf(stderr, "usage: %s %s\n", pname, usage);
    exit(1);
  }

