/* hr2test --- test program for hr2.h  */
/* Last edited on 2006-12-02 16:48:21 by stolfi */

#include <hr2.h>

#include <r2.h>
#include <r2x2.h>
#include <r3.h>
#include <r3x3.h>

#include <affirm.h>
#include <jsrandom.h>
#include <flt.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define NH 3
  /* Number of homogeneous coordinates in a point. */

#define NC 2
  /* Number of Cartesian coordinates in a point. */
 
#define check_eq(x, y, msg) \
  if ((x) != (y)) \
    { fprintf(stderr, " ** %20.16e %20.16e\n", (x), (y)); \
      affirm(0, msg); \
    } \
  else { }

#define check_eps(x, y, eps, msg) \
  if (fabs((x) - (y)) > (eps)) \
    { fprintf(stderr, " ** %20.16e %20.16e %20.16e\n", (x), (y), (eps)); \
      affirm(0, msg); \
    } \
  else { }

/* Internal prototypes */

int main (int argc, char **argv);
void test_hr2(int verbose);
void test_hr2_pmap(int verbose);
void throw_pmap(hr2_pmap_t *m);
double frac (double x);

int main (int argc, char **argv)
  {
    int i;
    srand(1993);
    srandom(1993);

    for (i = 0; i < 100; i++) test_hr2(i < 3);
    for (i = 0; i < 100; i++) test_hr2_pmap(i < 3);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }

double frac (double x)
  { int i = (int)x;
    double f = x - (double)i;
    return (f);
  }

void test_hr2(int verbose)
  {
    hr2_point_t p, q, r;
    hr2_line_t L, M, N;
    r2_t pc;
    int i;

    if (verbose)
      { fprintf(stderr,
          "sizeof(hr2_point_t) = %d  %d*sizeof(double) = %d\n",
          sizeof(hr2_point_t), NH, NH*sizeof(double)
        );
      }

    /* hr2_from_r2: */
    r2_throw_cube(&pc);
    p = hr2_from_r2(&pc);
    affirm(p.c.c[0] == 1.0, "hr2_from_r2 error(1)");
    
    /* r2_from_hr2: */
    r3_throw_cube(&(p.c));
    pc = r2_from_hr2(&p);
    q = hr2_from_r2(&pc);
    { double tol = fabs(p.c.c[0])*1.0e-12;
      for (i = 1; i < NC; i++)
        { double di = q.c.c[i]*p.c.c[0] - p.c.c[i];
          affirm(fabs(di) < tol, "r2_from_hr2 error(1)");
        }
    }
    
    /* hr2_side: */
    r3_throw_cube(&(p.c));
    r3_throw_cube(&(L.f));
    { double dd = r3_dot(&(p.c), &(L.f));
      sign_t sgn = hr2_side(&p, &L);
      affirm(((dd == 0.0) && (sgn == 0)) || (dd*sgn > 0), "hr2_side error(1)");
    }

    /* hr2_orient: */
    r3_throw_cube(&(p.c));
    r3_throw_cube(&(q.c));
    r3_throw_cube(&(r.c));
    { double dd = r3_det(&(p.c), &(q.c), &(r.c));
      sign_t sgn1 = hr2_orient(&p, &q, &r);
      affirm(((dd == 0.0) && (sgn1 == 0)) || (dd*sgn1 > 0), "hr2_orient error(1)");
      sign_t sgn2 = hr2_orient(&p, &p, &q);
      affirm(sgn2 == 0, "hr2_orient error(2)");
    }

    /* hr2_point_point_dir: */
    r3_throw_cube(&(p.c)); p.c.c[0] = fabs(p.c.c[0]);
    r3_throw_cube(&(q.c)); q.c.c[0] = fabs(q.c.c[0]);
    { r2_t pc = r2_from_hr2(&p);
      r2_t qc = r2_from_hr2(&q);
      r2_t upq = hr2_point_point_dir(&p, &q);
      r2_t vpq;
      r2_sub(&qc, &pc, &vpq);
      r2_dir(&vpq, &vpq);
      for (i = 0; i < NC; i++)
        { check_eps(upq.c[i], vpq.c[i], 1.0e-12, "hr2_point_point_dir error"); }
    }

    /* hr2_join: */
    r3_throw_cube(&(p.c));
    r3_throw_cube(&(q.c));
    r3_throw_cube(&(r.c)); /* Random test point */
    L = hr2_join(&p, &q);
    { double tp = r3_norm(&(L.f))*r3_norm(&(p.c))*1.0e-12;
      double dp = r3_dot(&(p.c), &(L.f));
      affirm(fabs(dp) < tp, "hr2_join error(1)");
      
      double tq = r3_norm(&(L.f))*r3_norm(&(q.c))*1.0e-12;
      double dq = r3_dot(&(q.c), &(L.f));
      affirm(fabs(dq) < tq, "hr2_join error(2)");
      
      sign_t sgnL = hr2_side(&r, &L);
      sign_t sgnO = hr2_orient(&p, &q, &r);
      affirm(sgnL == sgnO, "hr2_join error(3)");
    }

    /* hr2_line_dir: */
    r3_throw_cube(&(p.c)); p.c.c[0] = fabs(p.c.c[0]) + 0.00001;
    r3_throw_cube(&(q.c)); q.c.c[0] = fabs(q.c.c[0]) + 0.00001;
    L = hr2_join(&p, &q);
    { r2_t dL = hr2_line_dir(&L);
      r2_t eL = hr2_point_point_dir(&p, &q);;
      double tol = 1.0e-12;
      for (i = 0; i < NC; i++)
        { check_eps(dL.c[i], eL.c[i], tol, "hr2_line_dir error"); }
    }
      
    /* hr2_meet: */
    r3_throw_cube(&(L.f)); p.c = L.f;
    r3_throw_cube(&(M.f)); q.c = M.f;
    r = hr2_meet(&L, &M);
    N = hr2_join(&p, &q);
    for (i = 0; i < NH; i++)
      { check_eq(r.c.c[i], N.f.c[i], "hr2_meet error(1)"); }
      
    /* hr2_line_normal: */
    r3_throw_cube(&(L.f));
    { r2_t nL = hr2_line_normal(&L);
      r2_t mL = (r2_t){{L.f.c[1], L.f.c[2]}};
      r2_dir(&mL, &mL);
      double tol = 1.0e-12;
      for (i = 0; i < NC; i++)
        { check_eps(nL.c[i], mL.c[i], tol, "hr2_line_normal error"); }
    }
  }

void test_hr2_pmap(int verbose)
  {
    // hr2_pmap_t A, B, C;
    // hr2_point_t p, q, r;
    // int i, j, k;

    /* Size: */
    if (verbose)
      { fprintf(stderr,
          "sizeof(hr2_pmap_t) = %d  2*%d*%d*sizeof(double) = %d\n",
          sizeof(hr2_pmap_t), NH, NH, 2*NH*NH*sizeof(double)
        );
      }

    /* TO BE COMPLETED !!! */
    
    // /* r2x2_map_row, r2x2_map_col: */
    // throw_matrix(&A);
    // r2_throw_cube(&a);
    // r2x2_map_row(&a, &A, &b);
    // r2x2_map_col(&A, &a, &c);
    // r2_zero(&bb);
    // r2_zero(&cc);
    // for (i = 0; i < N; i++)
    //   { for (j = 0; j < N; j++)
    //       { bb.c[j] += a.c[i] * A.c[i][j];
    //         cc.c[i] += A.c[i][j] * a.c[j];
    //       }
    //   }
    // r = r2_dist(&b, &bb);
    // affirm(r < 0.000000001 * r2_norm(&bb), "r2_map_row error");
    // s = r2_dist(&c, &cc);
    // affirm(s < 0.000000001 * r2_norm(&cc), "r2_map_col error");
    // 
    // /* r2x2_mul: */
    // throw_matrix(&A);
    // throw_matrix(&B);
    // r2x2_mul(&A, &B, &C);
    // for (i = 0; i < N; i++)
    //   { for (j = 0; j < N; j++)
    //       { double sum = 0.0;
    //         for (k = 0; k < N; k++) { sum += A.c[i][k]*B.c[k][j]; }
    //         check_eps(C.c[i][j], sum, 0.000000001 * fabs(sum),
    //           "r2x2_mul error"
    //         );
    //       }
    //   }
    // 
    // /* r2x2_det: */
    // throw_matrix(&A);
    // for (i = 0; i < N; i++)
    //   { int k = (i + 1) % N;
    //     for (j = 0; j < N; j++)
    //       { /* Check for linearity */
    //         r = drandom();
    //         A.c[i][j] = r;
    //         rr = r2x2_det(&A);
    // 
    //         s = drandom();
    //         A.c[i][j] = s;
    //         ss = r2x2_det(&A);
    // 
    //         t = drandom();
    //         A.c[i][j] = r*(1-t) + s*t;
    //         tt = r2x2_det(&A);
    //         mag = fabs(rr) + fabs(ss) + fabs(tt);
    //         check_eps(tt, rr*(1.0 - t) + ss*t, 000000001 * mag,
    //           "r2x2_det error(1)"
    //         );
    //       }
    // 
    //     /* Row swap test: */
    //     r = r2x2_det(&A);
    //     for (j = 0; j < N; j++)
    //       { double t = A.c[i][j]; A.c[i][j] = A.c[k][j]; A.c[k][j] = t; }
    //     rr = r2x2_det(&A);
    //     mag = fabs(r) + fabs(rr);
    //     check_eps(r, -rr, 000000001 * mag, "r2x2_det error(2)");
    // 
    //     /* Col swap test: */
    //     r = r2x2_det(&A);
    //     for (j = 0; j < N; j++)
    //       { double t = A.c[j][i]; A.c[j][i] = A.c[j][k]; A.c[j][k] = t; }
    //     rr = r2x2_det(&A);
    //     mag = fabs(r) + fabs(rr);
    //     check_eps(r, -rr, 000000001 * mag, "r2x2_det error(3)");
    //   }
    // 
    // /* r2x2_inv: */
    // throw_matrix(&A);
    // r2x2_inv(&A, &B);
    // r2x2_mul(&A, &B, &C);
    // for (i = 0; i < N; i++)
    //   { for (j = 0; j < N; j++)
    //       { double val = (i == j ? 1.0 : 0.0);
    //         affirm((C.c[i][j] - val) < 000000001, "r2x2_inv error");
    //       }
    //   }
    // 
    // /* r2x2_print: */
    // if (verbose)
    //   { throw_matrix (&A);
    //     fprintf(stderr, "A = ");
    //     r2x2_print(stderr, &A);
    //     fputc('\n', stderr);
    //   }
  }  

void throw_pmap(hr2_pmap_t *m)
  {
    int i, j;
    r3_t a;
    for (i = 0; i < NH; i++)
      { r3_throw_cube(&a);
        for (j = 0; j < NH; j++) { m->dir.c[i][j] = a.c[j]; }
      }
    r3x3_inv(&(m->dir), &(m->inv));
  }
