/* r3test --- test program for r3.h, r3x3.h  */
/* Last edited on 2007-01-14 02:27:29 by stolfi */

#include <r3.h>
#include <r3x3.h>

#include <affirm.h>
#include <jsrandom.h>
#include <flt.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define N 3

void check_eq(double x, double y, char *msg);
void check_i_eq(int i, double x, double y, char *msg);
void check_ij_eq(int i, int j, double x, double y, char *msg);
void check_eps(double x, double y, double eps, char *msg);

void check_eq(double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** %20.16e %20.16e\n", (x), (y));
        affirm(0, msg);
      }
  }

void check_i_eq(int i, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d] %20.16e %20.16e\n", (i), (x), (y));
        affirm(0, msg);
      }
  }

void check_ij_eq(int i, int j, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d,%d] %20.16e %20.16e\n", (i), (j), (x), (y));
        affirm(0, msg);
      }
  }

void check_eps(double x, double y, double eps, char *msg)
  { if (fabs(x - y) > eps)
      { fprintf(stderr, " ** %20.16e %20.16e %20.16e\n", (x), (y), (eps));
        affirm(0, msg);
      }
  }

/* Internal prototypes */

int main (int argc, char **argv);
void test_r3(int verbose);
void test_r3x3(int verbose);
void throw_matrix(r3x3_t *m);
double frac (double x);

int main (int argc, char **argv)
  { int i;
    srand(1993);
    srandom(1993);

    for (i = 0; i < 100; i++) test_r3(i < 3);
    for (i = 0; i < 100; i++) test_r3x3(i < 3);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }

double frac (double x)
  { int i = (int)x;
    double f = x - (double)i;
    return (f);
  }

void test_r3(int verbose)
  { r3_t a, b, c, d, e, para, perp;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    if (verbose)
      { fprintf(stderr,
          "sizeof(r3_t) = %d  %d*sizeof(double) = %d\n",
          sizeof(r3_t), N, N*sizeof(double)
        );
      }

    /* r3_zero: */
    r3_zero(&a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 0.0, "r3_zero error"); }

    /* r3_all: */
    r3_all(3.14, &a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 3.14, "r3_all error"); }

    /* r3_axis: */
    for (k = 0; k < N; k++)
      { r3_axis(k, &a);
        for (i = 0; i < N; i++)
          { check_eq(a.c[i], (i == k ? 1.0 : 0.0), "r3_axis error"); }
      }

    /* r3_throw_cube */
    r3_throw_cube(&a);
    for (i = 0; i < N; i++)
      { affirm(a.c[i] != a.c[(i+1)%N], "r3_throw error(1)"); 
        affirm(frac(a.c[i]*256.0) != 0.0, "r3_throw error(1)"); 
        affirm((a.c[i] > -1.0) && (a.c[i] < 1.0), "r3_throw error(2)"); 
      }

    /* r3_add: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_add(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] + b.c[i], "r3_add error"); }

    /* r3_sub: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_sub(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] - b.c[i], "r3_sub error"); }

    /* r3_neg: */
    r3_throw_cube(&a);
    r3_neg(&a, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], - a.c[i], "r3_neg error"); }

    /* r3_scale: */
    s = drandom();
    r3_throw_cube(&a);
    r3_scale(s, &a, &d);
    for (i = 0; i < N; i++)
      { double zi = s*a.c[i];
        check_eq(d.c[i], zi, "r3_scale error(1)");
      }

    /* r3_mix: */
    s = drandom();
    t = drandom();
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_mix(s, &a, t, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = s * a.c[i] + t * b.c[i];
        check_eq(d.c[i], ddi, "r3_mix error");
      }

    /* r3_mix_in: */
    s = drandom();
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    d = b;
    r3_mix_in(s, &a, &d);
    for (i = 0; i < N; i++)
      { double ddi = b.c[i] + s * a.c[i];
        check_eq(d.c[i], ddi, "r3_mix_in error");
      }

    /* r3_weigh: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_weigh(&a, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = a.c[i] * b.c[i];
        check_eq(d.c[i], ddi, "r3_weigh error");
      }

    /* r3_norm, r3_norm_sqr, r3_L1_inf_norm: */
    r3_throw_cube(&a);
    r = r3_norm(&a);
    s = r3_norm_sqr(&a);
    t = r3_L_inf_norm(&a);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < N; i++)
      { double ai = fabs(a.c[i]);
        ss += ai*ai; 
        if (ai > tt) { tt = ai; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r3_norm error");
    check_eps(s, ss, 0.000000001 * ss, "r3_norm_sqr error");
    check_eq(t, tt, "r3_L_inf_norm error");

    /* r3_dist, r3_dist_sqr, r3_L_inf_dist: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r = r3_dist(&a, &b);
    s = r3_dist_sqr(&a, &b);
    t = r3_L_inf_dist(&a, &b);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < N; i++)
      { double di = fabs(a.c[i] - b.c[i]);
        ss += di*di; 
        if (di > tt) { tt = di; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r3_dist error");
    check_eps(s, ss, 0.000000001 * ss, "r3_dist_sqr error");
    check_eq(t, tt, "r3_L_inf_dist error");

    /* r3_dir, r3_L_inf_dir: */
    r3_throw_cube(&a);
    r = r3_dir(&a, &b);
    s = r3_L_inf_dir(&a, &d);
    ss = r3_norm(&a);
    tt = r3_L_inf_norm(&a);
    for (i = 0; i < N; i++)
      { check_eps(b.c[i], a.c[i]/ss, 0.000000001 * ss, "r3_dir error");
        check_eps(d.c[i], a.c[i]/tt, 0.000000001 * tt, "r3_L_inf_dir error");
      }

    /* r3_dot, r3_cos, r3_sin: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r = r3_dot(&a, &b);
    s = r3_sin(&a, &b);
    t = r3_cos(&a, &b);
    mag = sqrt(r3_dot(&a,&a)*r3_dot(&b,&b));
    rr = 0.0;
    for (i = 0; i < N; i++) { rr += a.c[i]*b.c[i]; }
    tt = rr/(r3_norm(&a)*r3_norm(&b));
    check_eps(r, rr, 0.000000001 * mag, "r3_dot error(1)");
    check_eps(t, tt, 0.000000001, "r3_cos error(1)");
    d = a;
    r3_mix_in(-rr/r3_norm_sqr(&b), &b, &d);
    ss = r3_norm(&d)/r3_norm(&a);
    check_eps(s, ss, 0.000000001, "r3_sin error(1)");

    for (i = 0; i < N; i++)
      { r3_axis(i, &a);
        for (j = 0; j < N; j++)
          { r3_axis(j, &b);
            r = r3_dot(&a, &b);
            s = r3_sin(&a, &b);
            t = r3_cos(&a, &b);
            rr = (i == j ? 1.0 : 0.0);
            check_eq(r, rr, "r3_dot error(2)");
            check_eq(t, rr, "r3_dot error(2)");
            check_eq(s, 1.0 - rr, "r3_dot error(2)");
          }
      }

    /* r3_cross: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        int i2 = (i + 2) % N;
        int p;
        r3_axis(i0, &a);
        r3_axis(i1, &b);
        r3_cross(&a, &b, &d);
        r3_axis(i2, &e);
        for (p = 0; p < N; p++)
          { double ep = e.c[p];
            check_eq(d.c[p], ep, "r3_cross error(x)");
          }
      }
    /* Test on random vectors: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_cross(&a, &b, &d);
    mag = r3_norm(&a)*r3_norm(&b);
    r = r3_dot(&a, &d);
    check_eps(r, 0.0, 0.00000001 * mag*r3_norm(&a), "r3_cross error(1)");
    r = r3_dot(&b, &d);
    check_eps(r, 0.0, 0.00000001 * mag*r3_norm(&b), "r3_cross error(2)");

    /* r3_det: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        int i2 = (i + 2) % N;
        r3_axis(i0, &a);
        r3_axis(i1, &b);
        r3_axis(i2, &c);
        r = r3_det(&a, &b, &c);
        check_eq(r, 1.0, "r3_det error(2)");
      }
    /* Test on random vectors: */
    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r3_throw_cube(&c);
    r = r3_det(&a, &b, &c);
    r3_cross(&a, &b, &e);
    rr = r3_dot(&e, &c);
    mag = r3_norm(&a)*r3_norm(&b)*r3_norm(&c);
    check_eps(r, rr, 0.00000001 * mag, "r3_det error(1)");

    /* r3_decomp: */

    r3_throw_cube(&a);
    r3_throw_cube(&b);
    r = r3_decomp(&a, &b, &para, &perp);
    rr = r3_dot(&a, &b)/r3_norm_sqr(&b);  
    check_eps(r, rr, 0.000000001 * (fabs(r) + fabs(rr)), "r3_decomp error(1)");
    r3_add(&para, &perp, &c);
    s = r3_dist(&a, &c);
    affirm (s <= 0.000000001 * r3_norm(&a), "r3_decomp error(2)");
    s = r3_dot(&perp, &b);
    check_eps(s, 0.0, 0.000000001 * r3_norm(&b), "r3_decomp error(3)");
    t = r3_dot(&para, &perp);
    check_eps(t, 0.0, 0.000000001 * r3_norm(&a), "r3_decomp error(4)");

    /* r3_print: */
    if (verbose)
      { r3_throw_cube (&a);
        fprintf(stderr, "a = ");
        r3_print(stderr, &a);
        fputc('\n', stderr);
      }
  }

void test_r3x3(int verbose)
  {
    r3x3_t A, B, C;
    r3_t a, b, c, bb, cc;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    /* Size and allocation: */
    if (verbose)
      { fprintf(stderr,
          "sizeof(r3x3_t) = %d  %d*%d*sizeof(double) = %d\n",
          sizeof(r3x3_t), N, N, N*N*sizeof(double)
        );
        fprintf(stderr, "&B = %08x\n", (unsigned)&B);
        fprintf(stderr, "&A-&B = %d\n", ((unsigned)(&A))-((unsigned)(&B)));
        fprintf(stderr, "&B-&C = %d\n", ((unsigned)(&B))-((unsigned)(&C)));
        fprintf(stderr, "&(B.c) = %08x\n", (unsigned)&(B.c));
        fprintf(stderr, "B.c = %08x\n", (unsigned)(B.c));
        fprintf(stderr, "&(B.c[0]) = %08x\n", (unsigned)&(B.c[0]));
        fprintf(stderr, "B.c[0] = %08x\n", (unsigned)(B.c[0]));
        fprintf(stderr, "&(B.c[0][0]) = %08x\n", (unsigned)&(B.c[0][0]));
      }

    /* Indexing and addressing: */
    for (i = 0; i < N; i++)
      for (j = 0; j < N; j++)
        { double *Aij = &(A.c[i][j]); 
          affirm(Aij = ((double *)&A)+(N*i)+j, "r3x3_t indexing error");
        }

    /* r3x3_zero, r3x3_ident: */
    r3x3_zero(&A);
    r3x3_ident(&B);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { check_eq(A.c[i][j], 0.0, "r3x3_zero error"); 
            check_eq(B.c[i][j], (i == j ? 1.0 : 0.0), "r3x3_ident error");
          }
      }

    /* r3x3_map_row, r3x3_map_col: */
    throw_matrix(&A);
    r3_throw_cube(&a);
    r3x3_map_row(&a, &A, &b);
    r3x3_map_col(&A, &a, &c);
    r3_zero(&bb);
    r3_zero(&cc);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { bb.c[j] += a.c[i] * A.c[i][j];
            cc.c[i] += A.c[i][j] * a.c[j];
          }
      }
    r = r3_dist(&b, &bb);
    affirm(r < 0.000000001 * r3_norm(&bb), "r3_map_row error");
    s = r3_dist(&c, &cc);
    affirm(s < 0.000000001 * r3_norm(&cc), "r3_map_col error");

    /* r3x3_mul: */
    throw_matrix(&A);
    throw_matrix(&B);
    r3x3_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double sum = 0.0;
            for (k = 0; k < N; k++) { sum += A.c[i][k]*B.c[k][j]; }
            check_eps(C.c[i][j], sum, 0.000000001 * fabs(sum),
              "r3x3_mul error"
            );
          }
      }

    /* r3x3_det: */
    throw_matrix(&A);
    for (i = 0; i < N; i++)
      { int k = (i + 1) % N;
        for (j = 0; j < N; j++)
          { /* Check for linearity */
            r = drandom();
            A.c[i][j] = r;
            rr = r3x3_det(&A);

            s = drandom();
            A.c[i][j] = s;
            ss = r3x3_det(&A);

            t = drandom();
            A.c[i][j] = r*(1-t) + s*t;
            tt = r3x3_det(&A);
            mag = fabs(rr) + fabs(ss) + fabs(tt);
            check_eps(tt, (rr*(1 - t) + ss*t), 000000001 * mag,
              "r3x3_det error(1)"
            );
          }

        /* Row swap test: */
        r = r3x3_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[i][j]; A.c[i][j] = A.c[k][j]; A.c[k][j] = t; }
        rr = r3x3_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r3x3_det error(2)");

        /* Col swap test: */
        r = r3x3_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[j][i]; A.c[j][i] = A.c[j][k]; A.c[j][k] = t; }
        rr = r3x3_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r3x3_det error(3)");
      }

    /* r3x3_inv: */
    throw_matrix(&A);
    r3x3_inv(&A, &B);
    r3x3_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double val = (i == j ? 1.0 : 0.0);
            affirm((C.c[i][j] - val) < 000000001, "r3x3_inv error");
          }
      }

    /* r3x3_print: */
    if (verbose)
      { throw_matrix (&A);
        fprintf(stderr, "A = ");
        r3x3_print(stderr, &A);
        fputc('\n', stderr);
      }
  }  

void throw_matrix(r3x3_t *m)
  { int i, j;
    r3_t a;
    for (i = 0; i < N; i++)
      { r3_throw_cube(&a);
        for (j = 0; j < N; j++) { m->c[i][j] = a.c[j]; }
      }
  }

