/* 
 * File:   cfConvert.cpp
 * Author: Mario J. Martin <dominonurbs$gmail.com>
 *
 * Converts a decimal floating point to binary base 2.
 * There are 10 kind of people in the world. Those who understand binary and those who do not.
 */

#include "cf_definitions.hpp"


// macro for abs. It seems that gcc, there is no abs(double)
#define _abs_(x) ((x > 0) ? x : -x)

/** Converts a decimal number into a custom-precision floating point representation base 2. 
* For now, it is only able to convert numbers that are less than LONG_MAX (usually 2147483647) It depends on the type used for man_t). 
* e.g. 1e100 or 1e-100 will fail to convert.
*/
template< unsigned int M, unsigned int E >
static CustomFloat< M, E > cfConvert(const double value)
{
	exp_t exponent = 0;	// exponent
	register man_t mantissa;	// mantissa
	register double fval;			// fraccional value
	man_t decimal;			// fractional part

	man_t sign = (value >= 0) ? 0 : CustomFloat< M, E >::MASK_SIGN_BIT;	// sign bit

	fval = (value >= 0) ? value : -value;	// make the number positive

	// check if the number is out of the available precission
	//const man_t max_range = ~((~0) << (sizeof(man_t) * 8 - 1));
	man_t max_range = ~0;
	max_range <<= (sizeof(man_t) * 8 - 1);
	max_range = ~max_range;

	if (value > max_range){
		_cf_handle_error_ ("Number out of range! \n%g is greater than %d \nThe ability to convert big numbers is not yet implemented!", value, max_range);

		#ifdef CF_THROW_ERRORS
			throw ("out of range!");
		#endif
	}

	// convert the integral part
	mantissa = (man_t)fval;

	//normalize
	if (mantissa > CustomFloat< M, E >::MASK_MANTISSA){
		// the mantissa if greater than the precission that the mantissa can handle (then, the fractional part is below precision and it is not calculated)
		while ((mantissa >> 1) > CustomFloat< M, E >::MASK_MANTISSA){		

#ifdef CF_CHECK_OVERFLOW 
			// check overflow
			if (exponent >= CustomFloat< M, E >::MASK_EXPONENT){
				_cf_handle_error_ ("Overflow detected! %g ", value);
				#ifdef CF_THROW_ERRORS
					throw("Overflow detected!");
				#endif
			}
#endif
			mantissa >>= 1;
			exponent++;
		}
		
		
		// notice that after normalization, the most significant bit should be always 1, except for special numbers
		mantissa >>= 1;
		exponent++;
		

		return CustomFloat< M, E >(mantissa | sign, exponent);
	}

	// convert the fractional part
	decimal = 0;
	if (mantissa == 0){
		// When mantissa is cero, it requires to find the most significant value
		bool flag = true;
		while (flag){
			fval *= 2;

			if (-(--exponent) > CustomFloat< M, E >::MASK_EXPONENT)
				return CustomFloat< M, E >(0, 0);	// the number is below precission.

			if (_abs_(fval) >= 1)
				flag = false;
		}

		mantissa = 1;
	}

	// The mantissa is not cero
	fval -= mantissa;
	man_t upper_bit = 1;
	upper_bit <<= (sizeof(man_t) * 8 - 1);
	fval *= upper_bit;

	decimal = (man_t)_abs_(fval);

	// take the last bit
	fval -= decimal;
	decimal <<= 1;
	fval *= 2;
	if (fval >= 1)
		decimal |= 1;

	// swap the decimal
	decimal = cfBinarySwap(decimal);

	while (mantissa < CustomFloat< M, E >::MASK_MANTISSA_MOST_SIGNIFICANT_BIT){
		mantissa <<= 1;
		mantissa |= (decimal & 1);
		decimal >>= 1;
		exponent--;
	}

	return CustomFloat< M, E >(mantissa | sign, exponent);
}

/** Converts a custom-precision floating point representation base 2 to double. */
template< unsigned int M, unsigned int E >
static double cfConvert(const CustomFloat< M, E >& f)
{
	double d = 0;
	man_t m = (f.m & CustomFloat< M, E >::MASK_MANTISSA);
	int sign = ((f.m & CustomFloat< M, E >::MASK_SIGN_BIT) != 0) ? -1 : 1;

	if (m == 0 && f.e == 0)
		return 0;
	
	d = m * pow(2., f.e) * sign;
	return d;
}

